# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from collections import defaultdict
from .models import ACTIONS
from .signals import (
    queued,
    updated,
    completed,
    failed,
)


class OrderDispatcher(object):
    """
    use_taskqueue : call task using .apply_async(args, kwargs)
    """
    def __init__(self, use_taskqueue=False):
        self.use_taskqueue = use_taskqueue
        self._handlers = defaultdict(dict)

    def register(self, resource_name, task, signame='queued', action='create',
                 queue_name=None):
        if action not in dict(ACTIONS):
            raise Exception('Unsupported action: {}'.format(action))

        if resource_name not in self._handlers[signame]:
            self._handlers[signame][resource_name] = {}

        if action not in self._handlers[signame][resource_name]:
            self._handlers[signame][resource_name][action] = []

        self._handlers[signame][resource_name][action].append((task, queue_name))

    def register_create(self, *args, **kwargs):
        return self.register(action='create', *args, **kwargs)

    def register_update(self, *args, **kwargs):
        return self.register(action='update', *args, **kwargs)

    def register_destroy(self, *args, **kwargs):
        return self.register(action='destroy', *args, **kwargs)

    def _call_tasks(self, signame, **kwargs):
        resource_name = kwargs.get('resource_name')
        context = kwargs.get('context') or {}
        uuid = kwargs.get('uuid')
        action = kwargs.get('action', 'create')

        if signame not in self._handlers:
            return

        if resource_name not in self._handlers[signame]:
            return

        if action not in self._handlers[signame][resource_name]:
            return

        task_list = self._handlers[signame][resource_name][action]

        for task, queue_name in task_list:
            if self.use_taskqueue:
                delayed_func = getattr(task, 'apply_async', None)
                if delayed_func is None:
                    raise RuntimeError(
                        'Task {} does not support .apply_async'.format(
                            repr(task))
                    )
                else:
                    if queue_name is not None:
                        task.apply_async(args=[uuid], kwargs=context, queue=queue_name)
                    else:
                        task.apply_async(args=[uuid], kwargs=context)
            else:
                task(uuid, **context)

    def process_queued(self, sender, **kwargs):
        self._call_tasks('queued', **kwargs)

    def process_updated(self, sender, **kwargs):
        self._call_tasks('updated', **kwargs)

    def process_completed(self, sender, **kwargs):
        self._call_tasks('completed', **kwargs)

    def process_failed(self, sender, **kwargs):
        self._call_tasks('failed', **kwargs)


dispatcher = OrderDispatcher()

queued.connect(dispatcher.process_queued)
updated.connect(dispatcher.process_updated)
completed.connect(dispatcher.process_completed)
failed.connect(dispatcher.process_failed)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from django.utils.translation import ugettext_lazy as _
import uuid
from django.db import models
from jsonfield import JSONField
import logging
from rest_framework.reverse import reverse
from .validators import validate_json
from .signals import (
    updated,
    queued,
    failed
)
from .constants import (
    ACTIONS,
    STATES,
)


class AbstractOrder(models.Model):
    class Meta:
        abstract = True

    uuid = models.UUIDField(default=uuid.uuid4, unique=True)

    status = models.CharField(max_length=32, choices=STATES, default='queued')

    resource_url = models.URLField(max_length=1024, default='')

    message = models.TextField(default='')

    context = JSONField(blank=True, validators=[validate_json])

    resource_name = models.CharField(max_length=200)

    action = models.CharField(max_length=32, choices=ACTIONS, default='create')

    queue_name = models.CharField(max_length=150, blank=True)

    created = models.DateTimeField(
        verbose_name=_('Created'),
        auto_now_add=True,
    )

    updated = models.DateTimeField(
        verbose_name=_('Updated'),
        auto_now=True,
    )

    progress = models.FloatField(default=0.0, null=True, blank=True)

    def get_absolute_url(self, request=None):
        return reverse(
            'async-rest:order-detail',
            args=[self.uuid],
            request=request,
        )

    def queue(self):
        """
        Queue order in task list
        """
        res = queued.send_robust(
            self.__class__,
            uuid=self.uuid,
            status=self.status,
            resource_url=self.resource_url,
            message=self.message,
            context=self.context,
            action=self.action,
            resource_name=self.resource_name,
        )
        return res

    def fail(self):
        res = failed.send_robust(
            self.__class__,
            uuid=self.uuid,
            status=self.status,
            message=self.message,
            resource_name=self.resource_name,
            context=self.context,
            resource_url=self.resource_url,
            progress=self.progress,
        )
        return res

    def save(self, *args, **kwargs):
        updated.send_robust(
            self.__class__,
            uuid=self.uuid,
            status=self.status,
            message=self.message,
            resource_name=self.resource_name,
            context=self.context,
            resource_url=self.resource_url,
            progress=self.progress,
        )

        return super(AbstractOrder, self).save(*args, **kwargs)

    def update_progress(self, percent):
        """
        Low level method to update progress
        """
        if percent >= 100.0:
            percent = 100.0

        if percent <= 0.0:
            percent = 0.0

        self.progress = percent
        self.save()

    def __unicode__(self):
        return '{} <{}:{}>'.format(self._meta.object_name, self.uuid, self.status)


class Order(AbstractOrder):
    pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from django.dispatch import Signal


attrs = ['uuid', 'resource_name', 'action', 'context', 'resource_url', 'message']

queued = Signal(providing_args=attrs)
updated = Signal(providing_args=attrs)
completed = Signal(providing_args=attrs)
failed = Signal(providing_args=attrs)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from django.core.exceptions import ValidationError


def validate_json(value):
    if type(value) is not dict:
        raise ValidationError('{} is not a valid JSON string.'.format(value))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from rest_framework import status
from rest_framework import test
from rest_framework.reverse import reverse
from async_rest.models import Order
from async_rest.dispatcher import dispatcher
from async_rest.helpers import fail_on_error


class Handler(object):
    def __init__(self, testname):
        self.task_call_count = {}
        self.task_call_count[testname] = 0
        self.testname = testname

    def task(self, uuid, **kwargs):
        self.task_call_count[self.testname] += 1

    def failed_task(self, uuid, **kwargs):
        self.task_call_count[self.testname] += 1

    @property
    def count(self):
        return self.task_call_count[self.testname]


class BasicTestSuite(test.APITestCase):
    def setUp(self):
        dispatcher.use_taskqueue = False

    def test_order_french_fries(self):
        context = {'size': 'XL'}
        resource_name = 'french-fries'

        handler = Handler('test_order_french_fries')
        dispatcher.register('french-fries', handler.task, 'queued')

        # Place an order
        res = self.client.post(
            reverse('async-rest:place-order',
                    kwargs={'resource_name': resource_name}),
            {
                'context': context
            },
            format='json'
        )
        self.assertEqual(res.status_code, status.HTTP_302_FOUND)


        # Retrieve order and get status

        res = self.client.get(res.url)
        self.assertEqual(res.status_code, status.HTTP_202_ACCEPTED)
        order = Order.objects.get(uuid=res.data['uuid'])

        # Test assertions
        self.assertEqual(res.data['status'], 'queued', msg=res.data['message'])
        self.assertEqual(res.data['context'], context)
        self.assertEqual(order.resource_name, resource_name)
        self.assertEqual(res.status_code, status.HTTP_202_ACCEPTED)

        # Proceed with order - Out of band modifications
        order.status = 'running'
        order.message = 'cooking...'
        order.save()

        # Retrieve order and get status
        res = self.client.get(order.get_absolute_url())

        # Test assertions
        self.assertEqual(res.data['status'], 'running')
        self.assertEqual(res.data['message'], 'cooking...')
        self.assertEqual(res.data['context'], context)
        self.assertEqual(order.resource_name, resource_name)
        self.assertEqual(res.status_code, status.HTTP_202_ACCEPTED)

        # Proceed with order - Out of band modifications
        order.status = 'completed'
        order.message = 'ready'
        order.resource_url = '/fries/bucket/2/'
        order.save()

        # Retrieve order and get status
        res = self.client.get(order.get_absolute_url())

        # Test assertions
        self.assertEqual(res.data['status'], 'completed')
        self.assertEqual(res.data['message'], 'ready')
        self.assertEqual(res.data['resource_url'], 'http://testserver/fries/bucket/2/')
        self.assertEqual(res.data['context'], context)
        self.assertEqual(order.resource_name, resource_name)
        self.assertEqual(res.status_code, status.HTTP_201_CREATED)

    def test_dispatcher(self):
        context = {
            'size': 'XXL'
        }
        resource_name = 'french-fries'

        handler = Handler('test_dispatcher')
        dispatcher.register('french-fries', handler.task, 'queued')

        # Place an order
        res = self.client.post(
            reverse('async-rest:place-order', kwargs={'resource_name':resource_name}),
            {
                'context': context
            },
            format='json'
        )
        self.assertEqual(handler.count, 1)

    def test_invalid_values(self):
        resource_name = 'rotted-french-fries'
        handler = Handler('test_invalid_values')

        def made_to_fail(order_uuid, **kwargs):
            order = Order.objects.get(uuid=order_uuid)
            with fail_on_error(order, 'msg', status='failed'):

                raise Exception('I was made to fail')

        # Register invalid call back (logged through logging/error)
        dispatcher.register(resource_name, made_to_fail, 'queued')
        dispatcher.register(resource_name, handler.failed_task, 'failed')

        # Place an invalid order
        res = self.client.post(
            reverse('async-rest:place-order', kwargs={'resource_name':resource_name}),
            {
                'context': {}
            },
            format='json'
        )
        self.assertEqual(res.status_code, status.HTTP_302_FOUND)
        # Retrieve order and get status

        res = self.client.get(res.url)
        # Test assertions
        self.assertEqual(res.data['status'], 'failed')
        self.assertEqual(handler.count, 1)

        # Place another invalid order
        res = self.client.post(
            reverse('async-rest:place-order', kwargs={'resource_name':resource_name}),
            {
                'context': '"size": "XL"'
            },
            format='json'
        )
        # Retrieve error
        context = res.data.get('context')
        self.assertIsNotNone(context)
        self.assertTrue(context[0].endswith('is not a valid JSON string.'))


    def test_actions(self):
        """
        Test multiple actions on order
        """
        resource_name = 'multiple-actions'
        handler_create = Handler('test_creation')
        handler_destroy = Handler('test_deletion')

        # Register 2 tasks on same resource
        # but with different actions
        dispatcher.register_create(resource_name, handler_create.task)
        dispatcher.register_destroy(resource_name, handler_destroy.task)

        # Common url
        url = reverse('async-rest:place-order', kwargs={'resource_name':resource_name})

        # No actions yet
        self.assertEqual(handler_create.count, 0)
        self.assertEqual(handler_destroy.count, 0)

        # Place a creation order
        res = self.client.post(
            url,
            {
                'context': {}
            },
            format='json'
        )
        self.assertEqual(res.status_code, status.HTTP_302_FOUND)
        # Retrieve order and get status

        res = self.client.get(res.url)

        # Check response
        self.assertEqual(res.status_code, status.HTTP_202_ACCEPTED)
        self.assertEqual(res.data['status'], 'queued')

        # One creation, no deletion yet
        self.assertEqual(handler_create.count, 1)
        self.assertEqual(handler_destroy.count, 0)

        # Place a deletion order
        res = self.client.delete(
            url,
            {
                'context': {}
            },
            format='json'
        )

        # Retrieve order and get status
        self.assertEqual(res.status_code, status.HTTP_302_FOUND)

        res = self.client.get(res.url)

        # Check response
        self.assertEqual(res.status_code, status.HTTP_202_ACCEPTED)
        self.assertEqual(res.data['status'], 'queued')

        # One creation, One deletion
        self.assertEqual(handler_create.count, 1)
        self.assertEqual(handler_destroy.count, 1)

    def tearDown(self):
        pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from rest_framework import serializers
from .models import Order


class OrderSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Order
        fields = (
            'uuid',
            'status',
            'message',
            'resource_url',
            'context',
            'progress',
        )

        read_only_fields = (
            'uuid',
            'status',
            'message',
            'resource_url',
            'progress',
        )

    resource_url = serializers.SerializerMethodField()

    def get_resource_url(self, instance):
        if instance is None:
            return ''

        return ''.join([
            self.context.get('request').scheme,
            '://',
            self.context.get('request').get_host(),
            instance.resource_url,
        ])



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from async_rest.models import Order
from async_rest.helpers import ProgressItems
from rest_framework import test


class ProgressTestSuite(test.APITestCase):
    def setUp(self):
        pass

    def test_progress(self):
        """
        Test the Progress helper
        """
        order = Order.objects.create()
        self.assertEqual(order.progress, 0.0)

        check = 0
        with ProgressItems(order, range(10)) as items:
            for item in items:
                self.assertEqual(item, check)
                self.assertEqual(Order.objects.first().progress, check*10.0)
                check += 1

    def test_progress_iterator(self):
        """
        Test the Progress helper
        using an iterator
        """
        length = 1000
        order = Order.objects.create()
        self.assertEqual(order.progress, 0.0)

        check = 0
        with ProgressItems(order, xrange(length), length=length) as items:
            for item in items:
                self.assertEqual(item, check)
                self.assertAlmostEqual(Order.objects.first().progress, check*0.1)
                check += 1



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function


default_app_config = 'async_rest.apps.Config'
VERSION = (1, 0, 0)


def get_version(version=None):
    return '.'.join(map(str, VERSION))


__version__ = get_version(VERSION)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function


ACTIONS = (
    ('create', 'Create a resource'),
    ('update', 'Update a resource'),
    ('destroy', 'Destroy a resource'),
)

STATES = (
    ('queued', 'queued'),
    ('running', 'running'),
    ('completed', 'completed'),
    ('failed', 'failed'),
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from django.utils.translation import ugettext_lazy as _
from django.apps import AppConfig


class Config(AppConfig):
    name = 'async_rest'
    verbose_name = _('Async REST')




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
from django.conf.urls import url, patterns
from .views import (
    PlaceOrder,
    OrderView,
)


urlpatterns = patterns(
    '',
    url(r'^(?P<resource_name>[^/]+)/order/$', PlaceOrder.as_view(),
        name='place-order'),
    url(r'^orders/(?P<uuid>[^/]+)/$', OrderView.as_view(),
        name='order-detail'),
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import jsonfield.fields
import uuid
import async_rest.validators


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Order',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('uuid', models.UUIDField(default=uuid.uuid4, unique=True)),
                ('status', models.CharField(default='queued', max_length=32, choices=[('queued', 'queued'), ('running', 'running'), ('completed', 'completed'), ('failed', 'failed')])),
                ('resource_url', models.URLField(default='', max_length=1024)),
                ('message', models.TextField(default='')),
                ('context', jsonfield.fields.JSONField(default=dict, blank=True, validators=[async_rest.validators.validate_json])),
                ('resource_name', models.CharField(max_length=200)),
                ('action', models.CharField(default='create', max_length=32, choices=[('create', 'Create a resource'), ('update', 'Update a resource'), ('destroy', 'Destroy a resource')])),
                ('queue_name', models.CharField(max_length=150, blank=True)),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='Created')),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='Updated')),
                ('progress', models.FloatField(default=0.0, null=True, blank=True)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, print_function
import StringIO
import traceback
import contextlib
import django.db
import logging


@contextlib.contextmanager
def fail_on_error(order, msg, status='failed'):
    """
    Context manager catching exception and failing automatically an order

    Logs error (traceback) using logging module and set ``message`` attribute
    of the order.
    """
    try:

        yield

    except Exception:
        msg_fp = StringIO.StringIO()
        traceback.print_exc(file=msg_fp)

        with django.db.transaction.atomic():
            order.message = '\n\nDetails:\n'.join([msg, msg_fp.getvalue()])
            order.status = status
            order.save()
            order.fail()

        message = 'Order {} failed.\n\n{}'.format(order.uuid, order.message)
        logging.error(message)
        raise RuntimeError(message)

    finally:
        order.save()


class ProgressItems(object):
    def __init__(self, order, items, length=None):
        self.order = order
        self.items = items
        if length and isinstance(length, int):
            self.length = length
        else:
            self.length = len(self.items)

    def __enter__(self):
        self.order.update_progress(0.0)

        for i, item in enumerate(self.items):
            yield item
            if self.length > 0:
                self.order.update_progress((i+1) * 100.0 / self.length)

    def __exit__(self, type, value, traceback):
        pass




# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages


ROOT = os.path.dirname(__file__)


def get_requirements(requirements_file):
    with open(requirements_file) as f:
        required = [line.split('#')[0] for line in f.read().splitlines()]
    return required


def extract_version():
    context = {}
    init_path = 'src/async_rest/__init__.py'
    with open(os.path.join(ROOT, init_path)) as fd:
        code = fd.read()
        exec(compile(code, init_path, "exec"), context)
    return context.get('__version__')


setup(
    name="django-async-rest",
    version=extract_version(),
    license='GPLv3',
    description="A RESTful way of dealing with asynchronous tasks",
    author='P.A. Schembri',
    author_email='pa.schembri@netsach.com',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://github.com/Netsach/django-async-rest',
    include_package_data=True,
    install_requires=get_requirements(os.path.join(ROOT, 'requirements.txt'))
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from django.db import transaction
from django.shortcuts import redirect
from rest_framework import status
from rest_framework import response
from rest_framework import generics
from ..serializers import OrderSerializer
from ..models import Order
import logging

logger = logging.getLogger()


class PlaceOrder(generics.CreateAPIView, generics.DestroyAPIView):
    model = Order
    serializer_class = OrderSerializer

    def build_order(self, request, action, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        with transaction.atomic():
            self.perform_create(serializer)
            serializer.instance.resource_name = kwargs.get('resource_name')
            serializer.instance.action = action
            serializer.instance.save()

        with transaction.atomic():
            order_uuid = serializer.instance.uuid
            order = Order.objects.get(uuid=order_uuid)

        order.queue()

        return redirect(order)

    def post(self, request, *args, **kwargs):
        return self.build_order(request, 'create', *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.build_order(request, 'update', *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        return self.build_order(request, 'destroy', *args, **kwargs)


class OrderView(generics.RetrieveAPIView):
    model = Order
    serializer_class = OrderSerializer
    queryset = Order.objects.all()
    lookup_field = 'uuid'

    def retrieve(self, request, *args, **kwargs):
        order = self.get_object()
        serializer = self.get_serializer(order)
        status_code = status.HTTP_200_OK

        if order.status in ['queued', 'running']:
            status_code = status.HTTP_202_ACCEPTED

        elif order.status == 'completed' and order.action == 'create':
            status_code = status.HTTP_201_CREATED

        return response.Response(serializer.data, status=status_code)
